from _typeshed import Incomplete
from enum import Enum
from pydantic_core import ErrorDetails as ErrorDetails
from slxjsonrpc.schema.jsonrpc import ErrorModel as ErrorModel, MethodError as MethodError, RpcBatch as RpcBatch, RpcError as RpcError, RpcErrorCode as RpcErrorCode, RpcErrorMsg as RpcErrorMsg, RpcNotification as RpcNotification, RpcRequest as RpcRequest, RpcResponse as RpcResponse, rpc_set_name as rpc_set_name, set_id_mapping as set_id_mapping, set_params_map as set_params_map, set_result_map as set_result_map
from typing import Any, Callable, Dict, Generator, List, Optional, Type, Union

RpcSchemas = Union[RpcError, RpcNotification, RpcRequest, RpcResponse]

class RpcErrorException(Exception):
    code: Incomplete
    msg: Incomplete
    data: Incomplete
    def __init__(self, code: Union[int, RpcErrorCode], msg: str, data: Optional[Any] = ...) -> None: ...
    def get_rpc_model(self, id: Union[str, int, None]) -> RpcError: ...

class SlxJsonRpc:
    log: Incomplete
    def __init__(self, methods: Optional[Enum] = ..., method_cb: Optional[Dict[Union[Enum, str], Callable[[Any], Any]]] = ..., result: Optional[Dict[Union[Enum, str], Union[type, Type[Any]]]] = ..., params: Optional[Dict[Union[Enum, str], Union[type, Type[Any]]]] = ...) -> None: ...
    def create_request(self, method: Union[Enum, str], callback: Callable[[Any], None], error_callback: Optional[Callable[[ErrorModel], None]] = ..., params: Optional[Any] = ...) -> Optional[RpcRequest]: ...
    def create_notification(self, method: Union[Enum, str], params: Optional[Any] = ...) -> Optional[RpcNotification]: ...
    def batch(self) -> Generator[None, None, None]: ...
    def batch_size(self) -> int: ...
    def get_batch_data(self, data: Optional[Union[RpcRequest, RpcNotification, RpcError, RpcResponse]] = ...) -> Optional[Union[RpcBatch, RpcRequest, RpcNotification, RpcError, RpcResponse]]: ...
    def parser(self, data: Union[bytes, str, Dict[str, Any], List[Dict[str, Any]]]) -> Optional[Union[RpcError, RpcResponse, RpcBatch]]: ...



"""The SlxJsonRpc Package."""
from slxjsonrpc.jsonrpc import SlxJsonRpc
from slxjsonrpc.jsonrpc import RpcErrorException

from slxjsonrpc.schema.jsonrpc import RpcBatch
from slxjsonrpc.schema.jsonrpc import RpcError
from slxjsonrpc.schema.jsonrpc import RpcNotification
from slxjsonrpc.schema.jsonrpc import RpcRequest
from slxjsonrpc.schema.jsonrpc import RpcResponse

__all__ = [
    'SlxJsonRpc',
    'RpcErrorException',
    'RpcBatch',
    'RpcError',
    'RpcNotification',
    'RpcRequest',
    'RpcResponse',
]

__version__ = "v0.9.0"
__auther__ = "Seluxit A/S"



from slxjsonrpc.jsonrpc import RpcErrorException as RpcErrorException, SlxJsonRpc as SlxJsonRpc
from slxjsonrpc.schema.jsonrpc import RpcBatch as RpcBatch, RpcError as RpcError, RpcNotification as RpcNotification, RpcRequest as RpcRequest, RpcResponse as RpcResponse



"""
Contains the JsonRpc Schemas used for the SlxJsonRpc Package.

The slxJsonRpc are build with the specification in mind, listed here:
    https://www.jsonrpc.org/specification
"""
import random
import string

from enum import Enum
from enum import IntEnum

from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field
from pydantic import field_validator
from pydantic import FieldValidationInfo
from pydantic import RootModel
from pydantic import TypeAdapter
from pydantic.fields import FieldInfo


from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union


_session_count: int = 0
_session_id: str = "".join(
    random.choices(string.ascii_letters + string.digits, k=10)
)

_RpcName: Optional[str] = None


def rpc_set_name(name: Optional[str]) -> None:
    """Set the JsonRpc id name."""
    global _RpcName
    _RpcName = name


def rpc_get_name() -> Optional[str]:
    """Retrieve the JsonRpc id name."""
    global _RpcName
    return _RpcName


def _id_gen(name: Optional[Union[str, int, float]] = None) -> str:
    """Create an unique Rpc-id."""
    global _session_count
    global _session_id
    global _RpcName
    rpc_name = name if name else _RpcName
    _session_count += 1
    return f"{_session_id}_{rpc_name}_{_session_count}"


class MethodError(Exception):
    """Exception to track if the Method is wrong."""
    pass


class RpcVersion(str, Enum):
    """The supported JsonRpc versions."""
    v2_0 = "2.0"


###############################################################################
#                             JsonRpc Request Object
###############################################################################

params_mapping: Dict[Union[Enum, str], Union[type, Type[Any]]] = {}


def set_params_map(mapping: Dict[Union[Enum, str], Union[type, Type[Any]]]) -> None:
    """Set the method to params schema mapping."""
    global params_mapping
    params_mapping = mapping


class RpcRequest(BaseModel):
    """
    The Standard JsonRpc Request Schema, used to do a request of the server.

    Attributes:
        jsonrpc: The JsonRpc version this schema is using. (Default v2.0)
        id: A identifier set by the client. (If emitted it will be auto generated)
        method: The name of the method to be invoked.
        params: (Optional) The input parameters for the invoked method.
    """
    jsonrpc: Optional[RpcVersion] = RpcVersion.v2_0
    method: str
    id: Optional[Union[str, int]] = Field(default=None, validate_default=True)
    params: Optional[Any] = Field(default=None, validate_default=True)

    """Enforce that there can not be added extra keys to the BaseModel."""
    model_config: ConfigDict = ConfigDict(extra='forbid')  # type: ignore

    @field_validator('id', mode='before')
    def id_autofill(cls, v: Optional[Union[str, int]], info: FieldValidationInfo) -> Union[str, int]:
        """Validate the id, and auto-fill it is not set."""
        return v or _id_gen(name=rpc_get_name() or info.data.get('method'))

    @classmethod
    def update_method(cls, new_type: Enum) -> None:
        """Update the Method schema, to fit the new one."""
        cls.model_fields['method'] = FieldInfo(
            annotation=new_type,
        )
        cls.__annotations__['method'] = new_type

    @field_validator("params")
    def method_params_mapper(cls, v: Optional[Any], info: FieldValidationInfo) -> Any:
        """Check & enforce the params schema, depended on the method value."""
        global params_mapping

        if not params_mapping.keys():
            return v

        if info.data.get('method') is None:
            # UNSURE: Why is this needed, when MethodError is use instead of ValueError? o.0
            return v

        if info.data.get('method') not in params_mapping.keys():
            raise MethodError(f"Unknown method: {info.data.get('method')}.")

        model = params_mapping[str(info.data.get('method'))]

        if isinstance(model, BaseModel):
            return model.model_validate(v)

        if model is not None:
            model_converter: TypeAdapter = TypeAdapter(model)  # type: ignore
            return model_converter.validate_python(v)

        if v:
            raise ValueError("params should not be set.")


class RpcNotification(BaseModel):
    """
    The Standard JsonRpc Notification Schema, to Notifies the server of change.

    Supposed to be a Request Object, just without the 'id'.

    Attributes:
        jsonrpc: The JsonRpc version this schema is using. (Default v2.0)
        method: The name of the method to be invoked.
        params: (Optional) The input parameters for the invoked method.
    """
    jsonrpc: Optional[RpcVersion] = RpcVersion.v2_0
    method: str
    params: Optional[Any] = Field(default=None, validate_default=True)

    """Enforce that there can not be added extra keys to the BaseModel."""
    model_config: ConfigDict = ConfigDict(extra='forbid')  # type: ignore

    @classmethod
    def update_method(cls, new_type: Enum) -> Any:
        """Update the Method schema, to fit the new one."""
        cls.model_fields['method'] = FieldInfo(
            annotation=new_type,
        )
        cls.__annotations__['method'] = new_type

    @field_validator("params")
    def method_params_mapper(cls, v: Optional[Any], info: FieldValidationInfo) -> Any:
        """Check & enforce the params schema, depended on the method value."""
        global params_mapping

        if not params_mapping.keys():
            return v

        if info.data.get('method') is None:
            # UNSURE: Why is this needed, when MethodError is use instead of ValueError? o.0
            return v

        if info.data.get('method') not in params_mapping.keys():
            raise MethodError(f"Unknown method: {info.data.get('method')}.")

        model = params_mapping[str(info.data.get('method'))]

        if isinstance(model, BaseModel):
            return model.model_validate(v)

        if model is not None:
            model_converter: TypeAdapter = TypeAdapter(model)  # type: ignore
            return model_converter.validate_python(v)

        if v:
            raise ValueError("params should not be set.")


###############################################################################
#                          JsonRpc Response Object
###############################################################################

result_mapping: Dict[Union[Enum, str], Union[type, Type[Any]]] = {}

id_mapping: Dict[Union[str, int, None], Union[Enum, str]] = {}


def set_id_mapping(mapping: Dict[Union[str, int, None], Union[Enum, str]]) -> None:
    """Set the id to method mapping."""
    global id_mapping
    id_mapping = mapping


def set_result_map(mapping: Dict[Union[Enum, str], Union[type, Type[Any]]]) -> None:
    """Set the method to params schema mapping."""
    global result_mapping
    result_mapping = mapping


class RpcResponse(BaseModel):
    """The Standard JsonRpc Response Schema, that is responded with.

    Attributes:
        jsonrpc: The JsonRpc version this schema is using. (Default v2.0)
        id: Must be the same value as the object this is a response to.
        result: The result of the Request object, if it did not fail.
    """
    jsonrpc: Optional[RpcVersion] = RpcVersion.v2_0
    id: Union[str, int]
    result: Any = Field(validate_default=True)

    """Enforce that there can not be added extra keys to the BaseModel."""
    model_config: ConfigDict = ConfigDict(extra='forbid')  # type: ignore

    @field_validator("result", mode='before')
    def method_params_mapper(cls, v: Any, info: FieldValidationInfo) -> Any:
        """Check & enforce the params schema, depended on the method value."""
        global result_mapping
        global method_id_mapping

        if not result_mapping.keys():
            return v

        the_id = info.data.get('id')

        if the_id not in id_mapping:
            # UNSURE (MBK): What should done, when it was not meant for this receiver?
            return v

        the_method = id_mapping[the_id]

        if the_method not in result_mapping.keys():
            raise ValueError(f"Not valid params for method: {info.data.get('method')}.")

        model = result_mapping[the_method]

        if isinstance(model, BaseModel):
            return model.model_validate(v)

        if model is not None:
            model_converter: TypeAdapter = TypeAdapter(model)  # type: ignore
            return model_converter.validate_python(v)

        if v:
            raise ValueError("result should not be set.")


###############################################################################
#                             JsonRpc Error Object
###############################################################################

class RpcErrorCode(IntEnum):
    """
    JsonRpc Standard Error Codes.

    Error Codes:    Error code:         Message Description:
    ---
        -32700      Parse error         Invalid JSON was received by the server.
                                        An error occurred on the server while parsing the JSON text.
        -32600      Invalid Request     The JSON sent is not a valid Request object.
        -32601      Method not found    The method does not exist / is not available.

        -32602      Invalid params      Invalid method parameter(s).
        -32603      Internal error      Internal JSON-RPC error.
        -32000      Server error        Internal error.
          ...
        -32099      Server error        Internal error.
    """
    ParseError = -32700
    InvalidRequest = -32600
    MethodNotFound = -32601
    InvalidParams = -32602
    InternalError = -32603
    ServerError = -32000


class RpcErrorMsg(str, Enum):
    """
    JsonRpc Standard Error Messages.

    Error Codes:    Error code:         Message Description:
    ---
        -32700      Parse error         Invalid JSON was received by the server.
                                        An error occurred on the server while parsing the JSON text.
        -32600      Invalid Request     The JSON sent is not a valid Request object.
        -32601      Method not found    The method does not exist / is not available.

        -32602      Invalid params      Invalid method parameter(s).
        -32603      Internal error      Internal JSON-RPC error.
        -32000      Server error        Internal error.
          ...
        -32099      Server error        Internal error.
    """
    ParseError = "Invalid JSON was received by the server."
    InvalidRequest = "The JSON sent is not a valid Request object."
    MethodNotFound = "The method does not exist / is not available."
    InvalidParams = "Invalid method parameter(s)."
    InternalError = "Internal JSON-RPC error."
    ServerError = "Internal server error."


class ErrorModel(BaseModel):
    """
    The Default JsonRpc Error message, that is responded with on error.

    Attributes:
        code: The error code.
        message: A short describing of the error.
        data: (Optional), a Additional information of the error.
    """
    code: Union[int, RpcErrorCode]
    message: str
    data: Optional[Any] = None

    """Enforce that there can not be added extra keys to the BaseModel."""
    model_config: ConfigDict = ConfigDict(extra='forbid')  # type: ignore

    @field_validator("code")
    def method_code_parser(
        cls,
        v: Union[str, bytes, int, float],
        info: FieldValidationInfo
    ) -> Union[int, RpcErrorCode]:
        """Error code parser."""
        value = int(v)

        if -32100 < value < -32000:
            return value

        return RpcErrorCode(value)


class RpcError(BaseModel):
    """
    The default JsonRpc Error Reply Schema.

    Attributes:
        jsonrpc:
        id:
        error:
    """
    id: Union[str, int, None] = None
    jsonrpc: Optional[RpcVersion] = RpcVersion.v2_0
    error: ErrorModel


###############################################################################
#                             JsonRpc Batch Object
###############################################################################

class RpcBatch(RootModel[List[Union[RpcRequest, RpcNotification, RpcResponse, RpcError]]]):
    """The Default JsonRpc Batch Schema."""
    root: List[Union[
        RpcRequest,
        RpcNotification,
        RpcResponse,
        RpcError,
    ]] = Field(..., min_length=1)



from enum import Enum, IntEnum
from pydantic import BaseModel, ConfigDict, FieldValidationInfo as FieldValidationInfo, RootModel
from typing import Any, Dict, List, Optional, Type, Union

def rpc_set_name(name: Optional[str]) -> None: ...
def rpc_get_name() -> Optional[str]: ...

class MethodError(Exception): ...

class RpcVersion(str, Enum):
    v2_0: str

params_mapping: Dict[Union[Enum, str], Union[type, Type[Any]]]

def set_params_map(mapping: Dict[Union[Enum, str], Union[type, Type[Any]]]) -> None: ...

class RpcRequest(BaseModel):
    jsonrpc: Optional[RpcVersion]
    method: str
    id: Optional[Union[str, int]]
    params: Optional[Any]
    model_config: ConfigDict  #type: ignore
    def id_autofill(cls, v: Optional[Union[str, int]], info: FieldValidationInfo) -> Union[str, int]: ...
    @classmethod
    def update_method(cls, new_type: Enum) -> None: ...
    def method_params_mapper(cls, v: Optional[Any], info: FieldValidationInfo) -> Any: ...

class RpcNotification(BaseModel):
    jsonrpc: Optional[RpcVersion]
    method: str
    params: Optional[Any]
    model_config: ConfigDict  #type: ignore
    @classmethod
    def update_method(cls, new_type: Enum) -> Any: ...
    def method_params_mapper(cls, v: Optional[Any], info: FieldValidationInfo) -> Any: ...

result_mapping: Dict[Union[Enum, str], Union[type, Type[Any]]]
id_mapping: Dict[Union[str, int, None], Union[Enum, str]]

def set_id_mapping(mapping: Dict[Union[str, int, None], Union[Enum, str]]) -> None: ...
def set_result_map(mapping: Dict[Union[Enum, str], Union[type, Type[Any]]]) -> None: ...

class RpcResponse(BaseModel):
    jsonrpc: Optional[RpcVersion]
    id: Union[str, int]
    result: Any
    model_config: ConfigDict  #type: ignore
    def method_params_mapper(cls, v: Any, info: FieldValidationInfo) -> Any: ...

class RpcErrorCode(IntEnum):
    ParseError: int
    InvalidRequest: int
    MethodNotFound: int
    InvalidParams: int
    InternalError: int
    ServerError: int

class RpcErrorMsg(str, Enum):
    ParseError: str
    InvalidRequest: str
    MethodNotFound: str
    InvalidParams: str
    InternalError: str
    ServerError: str

class ErrorModel(BaseModel):
    code: Union[int, RpcErrorCode]
    message: str
    data: Optional[Any]
    model_config: ConfigDict  #type: ignore
    def method_code_parser(cls, v: Union[str, bytes, int, float], info: FieldValidationInfo) -> Union[int, RpcErrorCode]: ...

class RpcError(BaseModel):
    id: Union[str, int, None]
    jsonrpc: Optional[RpcVersion]
    error: ErrorModel

class RpcBatch(RootModel[List[Union[RpcRequest, RpcNotification, RpcResponse, RpcError]]]):
    root: List[Union[RpcRequest, RpcNotification, RpcResponse, RpcError]]



from setuptools import find_packages
from setuptools import setup

import pathlib
import codecs
import os
import re


readme_file = pathlib.Path('README.md')
changelog_file = pathlib.Path('CHANGELOG.md')

here = os.path.abspath(os.path.dirname(__file__))


def get_long_description():
    long_description = ""

    if not readme_file.exists():
        return ""

    with readme_file.open("r") as file:
        long_description += file.read()

    long_description += "\n\n"

    if not changelog_file.exists():
        return long_description

    with changelog_file.open("r") as file:
        long_description += file.read()

    return long_description


def find_version(*file_paths):
    path = os.path.join(here, *file_paths)
    with codecs.open(path, 'r') as fp:
        version_file = fp.read()
        version_match = re.search(r"__version__ = ['\"]([^'\"]*)['\"]",
                                  version_file, re.M)
        if version_match:
            return version_match.group(1)
        raise RuntimeError("Unable to find version string.")


def find_auther(*file_paths):
    path = os.path.join(here, *file_paths)
    with codecs.open(path, 'r') as fp:
        auther_file = fp.read()
        auther_match = re.search(r"__auther__ = ['\"]([^'\"]*)['\"]",
                                 auther_file, re.M)
        if auther_match:
            return auther_match.group(1)
        raise RuntimeError("Unable to find auther string.")


setup(
    name="slxjsonrpc",
    version=find_version("slxjsonrpc", "__init__.py"),
    author=find_auther("slxjsonrpc", "__init__.py"),
    author_email="support@seluxit.com",
    license="Apache-2.0",
    description="SlxJsonRpc JsonRpc helper class, that uses pydantic.",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/Wappsto/slxjsonrpc",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ],
    packages=find_packages(),
    package_data={
        'slxjsonrpc': ["py.typed", "*.pyi", "**/*.pyi"],
    },
    tests_require=[
        'pytest',
        'tox'
    ],
    data_files=[('info', [readme_file.name, changelog_file.name])],
    install_requires=[
       'pydantic>=2.1.1'
    ],
    python_requires='>3.6.15',
)



# -*- coding: utf-8 -*-
"""Standalone JsonRpc module."""
import json
import logging

from contextlib import contextmanager

from typing import Any
from typing import Callable
from typing import Generator
from typing import Dict
from typing import List
from typing import Optional
from typing import overload
from typing import Type
from typing import Union

from enum import Enum

from pydantic import TypeAdapter
from pydantic import ValidationError
from pydantic_core import ErrorDetails

from slxjsonrpc.schema.jsonrpc import RpcBatch
from slxjsonrpc.schema.jsonrpc import RpcError
from slxjsonrpc.schema.jsonrpc import RpcNotification
from slxjsonrpc.schema.jsonrpc import RpcRequest
from slxjsonrpc.schema.jsonrpc import RpcResponse

from slxjsonrpc.schema.jsonrpc import ErrorModel
from slxjsonrpc.schema.jsonrpc import MethodError
from slxjsonrpc.schema.jsonrpc import rpc_set_name
from slxjsonrpc.schema.jsonrpc import RpcErrorCode
from slxjsonrpc.schema.jsonrpc import RpcErrorMsg
from slxjsonrpc.schema.jsonrpc import set_id_mapping
from slxjsonrpc.schema.jsonrpc import set_params_map
from slxjsonrpc.schema.jsonrpc import set_result_map

RpcSchemas = Union[
    RpcError,
    RpcNotification,
    RpcRequest,
    RpcResponse
]


class RpcErrorException(Exception):
    """
    Exception to reply a custom JsonRpc Error Response.

    This custom Exception extends the Exception class and implements
    a Rpc Error Code & message, to be transformed into the RpcError response.

    Attributes:
        Initializing with a msg & code arguments.
    """
    def __init__(
        self,
        code: Union[int, RpcErrorCode],
        msg: str,
        data: Optional[Any] = None
    ) -> None:
        """
        Initialize the RpcErrorException with the Rpc Error Response info.

        Args:
            code: The Rpc Error code, within the range of -32000 to -32099
            msg: The Rpc Error message, that shortly describe the error for given code.
            data: (Optional) The Rpc Extended error message.
        """
        super().__init__()
        self.code: Union[int, RpcErrorCode] = code
        self.msg: str = msg
        self.data: Optional[Any] = data

    def get_rpc_model(self, id: Union[str, int, None]) -> RpcError:
        """
        Returns a RpcError Response, for this given exception.

        The returned RpcError Model are used to send back to server,
        as a JsonRpc Error package.

        Args:
            id: The JsonRpc Id, for which this exception occurred.

        Returns:
            RpcError response fitting for this exception.
        """
        return RpcError(
            id=id,
            error=ErrorModel(
                code=self.code,
                message=self.msg,
                data=self.data
            )
        )


class SlxJsonRpc:
    """
    SlxJsonRpc is a JsonRpc helper class, that uses pydantic.

    SlxJsonRpc keep track of the JsonRpc schema, and procedure for each method.
    It also ensures to route each message to where it is expected.

    SlxJsonRpc is build to fill both the JsonRpc server & client roll.
    To enable the JsonRpc-server, the method_cb need to be given.


    """

    def __init__(
        self,
        methods: Optional[Enum] = None,
        method_cb: Optional[Dict[Union[Enum, str], Callable[[Any], Any]]] = None,
        result: Optional[Dict[Union[Enum, str], Union[type, Type[Any]]]] = None,
        params: Optional[Dict[Union[Enum, str], Union[type, Type[Any]]]] = None,
    ):
        """
        Initialization of the JsonRpc.

        # noqa: D417

        Args:
            method: (Optional) A String-Enum, with all the acceptable methods.
                    If not given, will there not be make checks for any wrong methods.
            method_cb: The mapping for each given method to a function call. (Server only)
                        callback: The function to be call when data is received.
                                  The Callback gets the params defined in as args,
                                  & should return the Result defined.
                                  If an error happens, and custom error code
                                  are needed, to send back, raise the RpcErrorException
                                  in the callback.
            result: (Optional) The method & 'result' mapping.
                    If not given, will there not be make checks for any wrong 'result'.
            params: (Optional) The Parser method & 'params' mapping.
                    If not given, will there not be make checks for any wrong 'params'.
        """
        self.log: logging.Logger = logging.getLogger(__name__)
        self.log.addHandler(logging.NullHandler())

        rpc_set_name(None)
        if methods:
            RpcRequest.update_method(methods)
            RpcNotification.update_method(methods)
        if params:
            set_params_map(params)
        if result:
            set_result_map(result)

        self.__batch_lock: int = 0
        self.__batched_list: List[RpcSchemas] = []

        self.__parse_as_rpc_obj: TypeAdapter = TypeAdapter(RpcSchemas)  # type: ignore

        self._method_cb: Dict[Union[Enum, str], Callable[[Any], Any]] = method_cb if method_cb else {}

        self._id_cb: Dict[Union[str, int, None], Callable[[Any], None]] = {}
        self._id_error_cb: Dict[Union[str, int, None], Callable[[Any], None]] = {}
        self._id_method: Dict[Union[str, int, None], Union[Enum, str]] = {}

        set_id_mapping(self._id_method)

    def create_request(
        self,
        method: Union[Enum, str],
        callback: Callable[[Any], None],
        error_callback: Optional[Callable[[ErrorModel], None]] = None,
        params: Optional[Any] = None,
    ) -> Optional[RpcRequest]:
        """
        Create a JsonRpc Request, with given method & params.

        The Created Request, are guaranteed to fit the given schema.
        When the Request are created, it will make sure that when the reply
        for the given request are received (through the parser-method),
        it will be passed on to the callback.

        Args:
            method: Should be a apart of the given Method Enum, given on init,
                    or if not given, a string.
            callback: The function to be called when data is received.
                      The Callback gets the Result data model (if set)
                      else a Dict/List back as argument.
            error_callback: (Optional) The function to be called, when an error
                            have happened.
                            The callback gets an ErrorModel object as parameter.
            params: (Optional) Should be fitting the a DataModel,
                    if given on init, else a valid Dictionary or List.

        Returns:
            RpcRequest, That should be send.

        Raises:
            ValidationError, if the given data do not fit the given Schema.
        """
        r_data = RpcRequest(
            method=method,
            params=params
        )

        self._add_result_handling(
            method=method,
            _id=r_data.id,
            callback=callback,
            error_callback=error_callback
        )

        return self._batch_filter(r_data)

    def _add_result_handling(
        self,
        method: Union[Enum, str],
        _id: Union[str, int, None],
        callback: Callable[[Any], None],
        error_callback: Optional[Callable[[ErrorModel], None]] = None,
    ) -> None:
        """Added handling for when a result comes back."""
        self._id_cb[_id] = callback
        if error_callback:
            self._id_error_cb[_id] = error_callback
        self._id_method[_id] = method

    def create_notification(
        self,
        method: Union[Enum, str],
        params: Optional[Any] = None,
    ) -> Optional[RpcNotification]:
        """
        Create a JsonRpc Notification, with given method & params.

        The Created Notification, are guaranteed to fit the given schema.
        Please note that there will not be a response for the notification
        send to the server.

        Args:
            method: Should be a apart of the given Method Enum, given on init,
                    or if not given, a string.
            params: (Optional) Should be fitting the a DataModel,
                    if given on init, else a valid Dictionary or List.

        Returns:
            The RPCNotification, to be send.

        Raises:
            ValidationError, if the given data do not fit the given Schema.
        """
        r_data = RpcNotification(
            method=method,
            params=params
        )

        return self._batch_filter(r_data)

    # -------------------------------------------------------------------------
    #                          Batching Functions
    # -------------------------------------------------------------------------

    @contextmanager
    def batch(self) -> Generator[None, None, None]:
        """Batch RPC's within the context manager, into one RPC-Batch-List."""
        self.__batch_lock += 1
        try:
            yield
        finally:
            self.__batch_lock -= 1

    def batch_size(self) -> int:
        """Retrieve the number of packages in the Bulk."""
        return len(self.__batched_list)

    def get_batch_data(
        self,
        data: Optional[Union[RpcRequest, RpcNotification, RpcError, RpcResponse]] = None
    ) -> Optional[Union[RpcBatch, RpcRequest, RpcNotification, RpcError, RpcResponse]]:
        """
        Retrieve the Bulked packages.

        The returned Package, will be a RpcBatch,
        unless it was empty, where it will return `None`, or
        it only contain one package, then that will be returned instead.

        Args:
            data: (Optional) If given the data are added to the end of the batched data.

        Returns:
            RpcBatch, if there was batch anything.
            None, if nothing was batch.
        """
        if len(self.__batched_list) < 1:
            return None
        if data:
            self.__batched_list.append(data)
        batched_data = self.__batched_list.copy()
        self.__batched_list.clear()
        if len(batched_data) == 1:
            return batched_data[0]
        return RpcBatch.model_validate(batched_data)

    @overload
    def _batch_filter(self, data: RpcError) -> Optional[RpcError]: ...  # noqa: E704

    @overload
    def _batch_filter(self, data: RpcNotification) -> Optional[RpcNotification]: ...  # noqa: E704

    @overload
    def _batch_filter(self, data: RpcResponse) -> Optional[RpcResponse]: ...  # noqa: E704

    @overload
    def _batch_filter(self, data: RpcRequest) -> Optional[RpcRequest]: ...  # noqa: E704

    def _batch_filter(
        self,
        data: RpcSchemas,
    ) -> Optional[RpcSchemas]:
        """
        Check if batch is enabled, and return the right reply.

        Args:
            data: RpcPackage to be returned if batch is not enabled.

        Returns:
            None, If Batch is enabled.
            data, if Batch is disabled.
        """
        if not self.__batch_lock:
            return data

        self.__batched_list.append(data)
        return None

    # -------------------------------------------------------------------------
    #                          Parsing Functions
    # -------------------------------------------------------------------------

    def parser(
        self,
        data: Union[bytes, str, Dict[str, Any], List[Dict[str, Any]]]
    ) -> Optional[Union[RpcError, RpcResponse, RpcBatch]]:
        """
        Parse raw JsonRpc data, & returns the Response or Error.

        For the Parsed data, there will be make a check for any method
        callbacks, if found, the callback(s) will be called for the given data,
        and the return value from the callback will be packed into, a jsonrpc
        response package, and returned here.

        Everything returned from this method, should be passed on to the
        receiver, if not `None`.

        Args:
            data: The Raw data to be parsed.

        Returns:
            The fitting JsonRpc reply to the given data.
            None, if no reply are needed.
        """
        try:
            j_data: Union[Dict[str, Any], List[Dict[str, Any]]]
            if isinstance(data, dict) or isinstance(data, list):
                j_data = data
            else:
                j_data = json.loads(data)

        except json.decoder.JSONDecodeError as err:
            return self._batch_filter(RpcError(
                id=None,
                error=ErrorModel(
                    code=RpcErrorCode.ParseError,
                    message=RpcErrorMsg.ParseError,
                    data=err.msg
                )
            ))

        if not j_data:
            return self._batch_filter(RpcError(
                id=None,
                error=ErrorModel(
                    code=RpcErrorCode.InvalidRequest,
                    message=RpcErrorMsg.InvalidRequest,
                )
            ))

        if isinstance(j_data, list):
            b_data: List[Union[RpcError, RpcResponse]] = []
            for f_data in j_data:
                try:
                    temp = self._parse_data(f_data)
                except RpcErrorException as err:
                    b_data.append(err.get_rpc_model(
                        id=f_data['id'] if 'id' in f_data else None,
                    ))
                    continue
                if temp:
                    r_data = self.__reply_logic(temp)
                    if r_data:
                        b_data.append(r_data)
                # UNSURE: Is it required to return a batch of 1, if it was received as batch of 1?

            return RpcBatch.model_validate(b_data) if b_data else None

        try:
            p_data = self._parse_data(j_data)
        except RpcErrorException as err:
            return self._batch_filter(err.get_rpc_model(
                id=j_data['id'] if 'id' in j_data else None,
            ))
        return self.__reply_logic(p_data)

    def __reply_logic(
        self,
        p_data: RpcSchemas
    ) -> Optional[Union[RpcResponse, RpcError]]:
        try:
            if isinstance(p_data, RpcError):
                return self._error_reply_logic(data=p_data)

            elif isinstance(p_data, RpcNotification):
                return self._notification_reply_logic(data=p_data)

            elif isinstance(p_data, RpcRequest):
                return self._request_reply_logic(data=p_data)

            elif isinstance(p_data, RpcResponse):
                return self._response_reply_logic(data=p_data)

        except RpcErrorException as err:
            return self._batch_filter(err.get_rpc_model(
                id=getattr(p_data, 'id', None),
            ))

        except Exception as err:
            print(f"Normal: {err}")  # TODO: Testing needed to trigger this!
            return self._batch_filter(RpcError(
                id=getattr(p_data, 'id', None),
                error=ErrorModel(
                    code=RpcErrorCode.InternalError,
                    message=RpcErrorMsg.InternalError,
                    data=err.args[0]
                )
            ))

        return None

    def _error_reply_logic(self, data: RpcError) -> Optional[RpcError]:
        if data.id not in self._id_cb.keys():
            # NOTE: Triggers only if it was an error that we generated.
            return data
        self._id_cb.pop(data.id)
        if data.id not in self._id_error_cb.keys():
            self.log.warning(f"Unhanded error: {data}")
        else:
            with self._except_handler():
                cb = self._id_error_cb.pop(data.id)
                self.log.debug(f"Exec Error CB: {cb}")
                cb(data.error)
        return None

    def _notification_reply_logic(self, data: RpcNotification) -> Optional[RpcError]:
        if data.method not in self._method_cb.keys():
            return self._batch_filter(RpcError(
                id=None,
                error=ErrorModel(
                    code=RpcErrorCode.MethodNotFound,
                    message=RpcErrorMsg.MethodNotFound,
                    data=data.method
                )
            ))
        with self._except_handler():
            cb = self._method_cb[data.method]
            self.log.debug(f"Exec Notification CB: {cb}")
            cb(data.params)
        return None

    def _request_reply_logic(self, data: RpcRequest) -> Optional[Union[RpcResponse, RpcError]]:
        if data.method in self._method_cb.keys():
            with self._except_handler():
                cb = self._method_cb[data.method]
                self.log.debug(f"Request CB: {cb}")
                result = cb(data.params)
            return self._batch_filter(RpcResponse(
                id=data.id,
                result=result
            ))
        return self._batch_filter(RpcError(
            id=data.id,
            error=ErrorModel(
                code=RpcErrorCode.MethodNotFound,
                message=RpcErrorMsg.MethodNotFound,
                data=data.method
            )
        ))

    def _response_reply_logic(self, data: RpcResponse) -> Optional[RpcResponse]:
        if data.id not in self._id_cb.keys():
            self.log.warning(f"Received an unknown RpcResponse: {data}")
        else:
            self._id_error_cb.pop(data.id, None)
            with self._except_handler():
                cb = self._id_cb.pop(data.id)
                self.log.debug(f"Exec Response CB: {cb}")
                cb(data.result)
        return None

    @contextmanager
    def _except_handler(self) -> Generator[None, None, None]:
        try:
            yield
        except RpcErrorException:
            raise
        except Exception as err:
            # NOTE: Only triggered from user given function, or if it was not a function
            self.log.exception("An error happened doing execution of a callback.")
            raise RpcErrorException(
                code=RpcErrorCode.ServerError,
                msg=RpcErrorMsg.ServerError,
                data=err.args[0]
            ).with_traceback(err.__traceback__)

    def __ValidationError2ErrorModel(
        self,
        errors: List[ErrorDetails]
    ) -> Optional[ErrorModel]:
        # TODO (MBK): Find a faster/better way to do this!
        params_error = list(filter(lambda x: x['loc'][1] == 'params', errors))
        type_error = list(filter(lambda x: x['type'] in ["missing", "extra_forbidden"], errors))
        if params_error:
            raise RpcErrorException(
                code=RpcErrorCode.InvalidParams,
                msg=RpcErrorMsg.InvalidParams,
                data=params_error[0]
            )
        elif type_error:
            raise RpcErrorException(
                code=RpcErrorCode.InvalidRequest,
                msg=RpcErrorMsg.InvalidRequest,
                data=type_error[0]
            )

        return None

    def _parse_data(
        self,
        data: Dict[str, Any]
    ) -> RpcSchemas:
        try:
            p_data: RpcSchemas = self.__parse_as_rpc_obj.validate_python(data)
        except MethodError as error:
            raise RpcErrorException(
                code=RpcErrorCode.MethodNotFound,
                msg=RpcErrorMsg.MethodNotFound,
                data=error.args[0]
            )
        except ValidationError as error:
            error_package = self.__ValidationError2ErrorModel(
                errors=error.errors()
            )

            if not error_package:
                # TODO: Testing needed to trigger this!
                self.log.exception(f"Unhanded ValidationError: {data}")
                raise

        return p_data
